import inspect
import warnings

from typing import Any, ClassVar, Dict, Optional, Tuple, Type, cast

import pydantic

from pydantic._internal._model_construction import (
    ModelMetaclass as PydanticModelMetaclass,
)
from pydantic._internal._model_construction import (
    PydanticGenericMetadata,  # type: ignore
)
from rest_framework import serializers  # type: ignore
from typing_extensions import Self, dataclass_transform

from drf_pydantic.base_serializer import DrfPydanticSerializer
from drf_pydantic.config import DrfConfigDict
from drf_pydantic.parse import create_serializer_from_model
from drf_pydantic.utils import get_attr_owner


@dataclass_transform(kw_only_default=True, field_specifiers=(pydantic.Field,))
class ModelMetaclass(PydanticModelMetaclass, type):
    def __new__(
        mcs,
        cls_name: str,
        bases: Tuple[Type[Any], ...],
        namespace: Dict[str, Any],
        __pydantic_generic_metadata__: Optional[PydanticGenericMetadata] = None,
        __pydantic_reset_parent_namespace__: bool = True,
        _create_model_module: Optional[str] = None,
        **kwargs: Any,
    ):
        cls = super().__new__(
            mcs,
            cls_name,
            bases,
            namespace,
            __pydantic_generic_metadata__,
            __pydantic_reset_parent_namespace__,
            _create_model_module,
            **kwargs,
        )

        # Set drf_config by merging properties with the following priority:
        # 1. Pydanitc model itself (cls)
        # 2. Parent class of the pydantic model (cls)
        # 3. Default values
        drf_config = DrfConfigDict(
            validate_pydantic=False,
            validation_error="drf",
            backpopulate_after_validation=True,
        )
        for base in cls.__mro__[:2][::-1]:
            drf_config.update(getattr(base, "drf_config", DrfConfigDict()))
        setattr(cls, "drf_config", drf_config)

        # Serializer was set by the user (is already present and is not inherited)
        if hasattr(cls, "drf_serializer") and cls is get_attr_owner(
            cls, "drf_serializer"
        ):
            drf_serializer = getattr(cls, "drf_serializer")
            if not inspect.isclass(drf_serializer):
                raise TypeError(
                    f"drf_serializer must be a class, check class {cls.__name__}"
                )
            if not issubclass(drf_serializer, serializers.Serializer):
                raise TypeError(
                    f"{drf_serializer.__name__} is not a valid type "
                    f"for drf_serializer. Check class {cls.__name__}"
                )
            if not issubclass(drf_serializer, DrfPydanticSerializer):
                warnings.warn(
                    (
                        f"custom drf_serializer on model {cls.__name__} "
                        f"should be replaced with an instace of "
                        f"drf_pydantic.DrfPydanticSerializer "
                        f"(currently is {drf_serializer.__name__})"
                    ),
                    UserWarning,
                )
            drf_serializer = cast(Type[DrfPydanticSerializer[Any]], drf_serializer)
            if getattr(drf_serializer, "_pydantic_model", cls) is not cls:
                warnings.warn(
                    (
                        f"_pydantic_model on model {cls.__name__} doesn't match "
                        f"expected class {cls.__name__}: "
                        f"{getattr(drf_serializer, '_pydantic_model').__name__}"
                    ),
                    UserWarning,
                )
            setattr(drf_serializer, "_pydantic_model", cls)
            if not hasattr(drf_serializer, "_drf_config"):
                setattr(drf_serializer, "_drf_config", drf_config)
        # Serializer not declared on cls directly (missing or inherited)
        else:
            setattr(
                cls,
                "drf_serializer",
                create_serializer_from_model(cls, drf_config),
            )

        return cls


class BaseModel(pydantic.BaseModel, metaclass=ModelMetaclass):
    # Populated by the metaclass or manually set by the user
    drf_serializer: ClassVar[Type[DrfPydanticSerializer[Self]]]
    drf_config: ClassVar[DrfConfigDict]



import inspect

from types import GenericAlias
from typing import TYPE_CHECKING, Any, Optional, Tuple, Type

if TYPE_CHECKING:
    from types import UnionType

    _UnionGenericAlias = UnionType
    _GenericAlias = GenericAlias
else:
    # Union syntax using pipe (e.g., int | str) is only available in Python 3.10+
    from typing import _GenericAlias, _UnionGenericAlias

    try:
        from types import UnionType
    except ImportError:  # pragma: no cover
        UnionType = None


def get_union_members(type_: Any) -> Optional[Tuple[Type[Any], ...]]:
    """
    Get union members from a union type.

    Parameters
    ----------
    type_ : Any
        Union type.

    Returns
    -------
    tuple[type, ...], optional
        Union members.
        None if type_ is not a union type.

    """
    if isinstance(type_, _UnionGenericAlias) or (  # type: ignore
        UnionType is not None and isinstance(type_, UnionType)  # type: ignore
    ):
        return type_.__args__
    return None


def is_scalar(type_: Type[Any]) -> bool:
    """
    Check if type is a scalar type.

    Scalar field is any field that is not a pydantic model (this would be nested field)
    or a collection field (e.g., list[int]).
    Examples of scalar fields: int, float, datetime, pydantic.EmailStr

    Parameters
    ----------
    type_ : type
        Type.

    Returns
    -------
    bool
        True if type is a scalar type.

    """
    return not isinstance(type_, GenericAlias) and not isinstance(type_, _GenericAlias)


def get_attr_owner(cls: Type[Any], attr: str) -> Type[Any]:
    """
    Return the class in the MRO that first declares a given attribute.

    Parameters
    ----------
    cls : Type[Any]
        Class whose MRO will be searched.
    attr : str
        Attribute name to locate within the MRO.

    Returns
    -------
    Type[Any]
        The class from the MRO that explicitly defines 'attr'.

    """
    if not hasattr(cls, attr):
        raise AttributeError(f"Class {cls.__class__.__name__} doesn't have {attr}")

    for base in inspect.getmro(cls):
        if attr in base.__dict__:
            return base

    # We know cls has attr, so this line should never be reached; for type checker
    raise RuntimeError  # pragma: no cover



from typing import Literal, TypedDict


class DrfConfigDict(TypedDict, total=False):
    validate_pydantic: bool
    """
    Whether to validate parent pydantic model on drf_serializer validation.

    By default is False.

    """
    validation_error: Literal["drf", "pydantic"]
    """
    What error to raise if pydantic model validation raises its ValidationError.

    !!! CAUTION !!!
    This will make the generated DRF serializer potentially
    incompatible with DRF views which expect DRF's ValidationError.

    By default 'drf'.
    'validate_pydantic' must be True for this to have any effect.

    """
    backpopulate_after_validation: bool
    """
    Update values in the DRF serializer after pydantic model validation.

    Useful if your pydantic validators modify model fields' values.

    By default True.
    'validate_pydantic' must be True for this to have any effect.

    """



import json
import warnings

from typing import Any, ClassVar, Dict, Generic, List, Optional, Type, TypeVar

import pydantic

from rest_framework import serializers  # type: ignore
from rest_framework.settings import api_settings  # type: ignore

from drf_pydantic.config import DrfConfigDict

T = TypeVar("T", bound=Dict[str, Any])
P = TypeVar("P", bound=pydantic.BaseModel)


class DrfPydanticSerializer(serializers.Serializer, Generic[P]):
    _pydantic_model: Type[P]
    _drf_config: ClassVar[DrfConfigDict]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)  # type: ignore
        self.__pydantic_instance: Optional[P] = None

    @property
    def pydantic_instance(self) -> P:
        if not self._drf_config.get("validate_pydantic"):
            raise AssertionError(
                "You must enable pydantic validation with `validate_pydantic` "
                "before accessing `.pydantic_instance`."
            )
        if self.__pydantic_instance is None:
            raise AssertionError(
                "You must call `.is_valid()` before accessing `.pydantic_instance`."
            )
        return self.__pydantic_instance

    def validate(self, attrs: T) -> T:
        return_value = super().validate(attrs)  # type: ignore
        if not self._drf_config.get("validate_pydantic"):
            return return_value

        try:
            validated_pydantic_model = self._pydantic_model(**attrs)
        except pydantic.ValidationError as exc:
            if self._drf_config.get("validation_error") == "pydantic":
                raise exc
            assert self._drf_config.get("validation_error") == "drf"

            field_errors: Dict[str, List[str]] = {}
            non_field_errors: List[str] = []
            for error in exc.errors():
                try:
                    message = json.dumps(
                        {
                            "loc": error["loc"],
                            "msg": error["msg"],
                            "type": error["type"],
                        }
                    )
                except Exception:  # pragma: no cover
                    message = f"{error['msg']} (type={error['type']})"
                if (
                    len(error["loc"]) == 0
                    or not isinstance(error["loc"][0], str)
                    or error["loc"][0] not in self.fields
                ):
                    non_field_errors.append(message)
                else:
                    field_errors.setdefault(error["loc"][0], []).append(message)

            raise serializers.ValidationError(
                {
                    getattr(
                        api_settings,
                        "NON_FIELD_ERRORS_KEY",
                        "non_field_errors",
                    ): non_field_errors,
                    **field_errors,
                }
            ) from exc

        if self._drf_config.get("backpopulate_after_validation"):
            for key in return_value:
                try:
                    pydantic_value = getattr(validated_pydantic_model, key)
                    if isinstance(pydantic_value, pydantic.BaseModel):
                        pydantic_value = pydantic_value.model_dump()
                    return_value[key] = pydantic_value
                except AttributeError:  # pragma: no cover
                    warnings.warn(
                        f"Failed to set attribute `{key}` when validating instance "
                        f"of type `{self._pydantic_model.__name__}`"
                    )
                    continue

        self.__pydantic_instance = validated_pydantic_model

        return return_value
